// Copyright (c) 1998-2002 B2C2, Incorporated.  All Rights Reserved.
//
// THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF B2C2, INCORPORATED.
// The copyright notice above does not evidence any
// actual or intended publication of such source code.
//
// This file is proprietary source code of B2C2, Incorporated. and is released pursuant to and
// subject to the restrictions of the non-disclosure agreement and license contract entered
// into by the parties.

//
// File: b2dump.cpp
//

//
// Change history:
//		11/18/2002: Changed code not to require PIDs to be set. If pids have already been set
//					(e.g. via b2settuner, etc) data will be received from those pids.
//



#include "b2dump.h"
#include "b2c2av.h"

// Command line usage message

#define	USAGE_MSG "\
\n\
Usage: b2dump arguments...\n\
\n\
Note: Arguments must be separated from values with blanks\n\
\n\
Required Arguments:\n\
-a    Adapter name; e.g. eth1\n\
-pt   TS PID as decimal or hex.; e.g. 164, 0xa4 or 0XA4.\n\
      For each TS PID, specify a separate -pt option;\n\
      e.g. -pt 17 -pt 18, etc.\n\
\n\
-pv   Video PID as decimal or hex.; e.g. 164, 0xa4 or 0XA4.\n\
      e.g. -pv 117 \n\
\n\
-pa   Audio PID as decimal or hex.; e.g. 164, 0xa4 or 0XA4.\n\
      e.g. -pa 118 \n\
\n\
Optional Arguments:\n\
-f    Name of file for data dump(default: \"b2dump.data.<TYPE>\" in current directory);\n\
      no quotes, no spaces. <TYPE> is \".TS\", \".AUD\" or \".VID\" and identifies the data type.\n\
      e.g. b2dump.data.ts, b2dump.data.vid.\n\
-t    Time duration for data dump in seconds(default: 10 sec.)\n\
-m    Monitor mode: don't write data to file, just report bytes received.\n\
-v    Verbose. \n\
-h    Help\n"


#define NO_PID_SET (-1L)

// Struct used for processing command line arguments.


FILE * log;

int	CheckCommandLineArgs(int argc, char * argv[], tSParams * pSParams);
int	check_and_write(unsigned char * m_pucBaseAddr, unsigned long offset,
		unsigned long len, unsigned short * m_usWrapCnt, FILE *);


B2C2MPEG2Adapter * CStreamRec::ms_pB2C2Adapter = NULL;
unsigned CStreamRec::ms_uInstanceCounter = 0;


CStreamRec::CStreamRec(eStreamType ThisStream, tSParams *pSParams) :
							m_ThisStream(ThisStream), m_pStream(NULL), m_pDump(NULL),
							m_ulLastPos(0), m_nOffset(0), m_lActivePid(NO_PID_SET)
{
	ms_uInstanceCounter++;
	m_lActivePid = -1;

	if (!InterfacesCreated())
	{
		// Emulate instantiating of DirectShow filter.
		ms_pB2C2Adapter = new B2C2MPEG2Adapter(pSParams->szAdapterName);
	}
}

CStreamRec::~CStreamRec ()
{
	int nRetVal = 0;
	HRESULT hr;

	// If PIDs were set, delete them.
	if (m_lActivePid != -1)
	{
		switch(m_ThisStream)
		{
			case TsStream:
				// !!!
				// Note: currently, this code only deletes the first TS pid, not the whole list.
				// !!!
//					hr = ms_pB2C2Adapter->GetDataControl()->DeleteTsPIDs((long)SParams.lTotalTSPidsSpecified,
//				                                     (long *)SParams.lTSPidArray);
				hr = ms_pB2C2Adapter->GetDataControl()->DeleteTsPIDs((long)1, &m_lActivePid);
				break;
			case VideoStream:
				hr = ms_pB2C2Adapter->GetAvControl()->DeleteAudioVideoPIDs(0, m_lActivePid);
				break;
			case AudioStream:
				hr = ms_pB2C2Adapter->GetAvControl()->DeleteAudioVideoPIDs(m_lActivePid, 0);
				break;
				
			default:
				fprintf(stderr,"\nb2dump: Unknown stream %d.\n", m_ThisStream);
				fflush(stderr);
				break;

		}

		if (FAILED (hr))
		{

			fprintf(stderr,"\nb2dump: Error: B2C2 Driver Data Ctrl."
					       " Deleting Pids method "
						   "failed, code: 0x%8X\n", hr);
			fflush(stderr);

			nRetVal = -1;
		}
	}


	if ((!--ms_uInstanceCounter) && InterfacesCreated()) 
	{

		delete ms_pB2C2Adapter;
		ms_pB2C2Adapter = NULL;
	}
}


int CStreamRec::StartStream(const tSParams &SParams, const CStreamRec * pPartner)
{
	int nRetVal = 0;
	HRESULT hr;
	char caName[1024];
	char *szSuffix;
	long lPartnerPid = 0;

	strcpy(caName,MAJOR_DEVICE_NAME);

	switch(m_ThisStream)
	{
		case TsStream:

			 if (SParams.bVerboseMode)
			 {
				fprintf(stderr,"\nb2dump: TsStream.");
			 }
             szSuffix = TS_SUFFIX;

			 if (SParams.lTotalTSPidsSpecified)
			 {

				hr = ms_pB2C2Adapter->GetDataControl()->
				           AddTsPIDs((long)SParams.lTotalTSPidsSpecified,
				                                     (long *)SParams.lTSPidArray);

				if (FAILED (hr)) 
				{

					fprintf(stderr,"\nb2dump: Error: B2C2 Driver Data Ctrl."
					               " Interface AddPIDs method failed, code: 0x%8X\n", hr);
					fflush(stderr);

					nRetVal = -1;

				} 
				else
				{				
					m_lActivePid = SParams.lTSPidArray[0];
				}
			 }

			 break;

		case VideoStream:

			if (SParams.bVerboseMode)
			{
				fprintf(stderr,"\nb2dump: VideoStream.");
			}
			szSuffix = VID_SUFFIX;

			 if (SParams.lTotalVidPidsSpecified)
			 {
			 	if (NULL != pPartner)
				    // Get the Audio Pid if there is one.
					lPartnerPid = pPartner->GetActivePid();

				if (SParams.bVerboseMode)
				{
					fprintf(stderr,"\nb2dump: StartStream(VideoStream) -- Audio Pid %04X,"
					        " Video Pid %04X.\n", lPartnerPid,
							(long)SParams.lVidPidArray[0]);
				}

				// To set only one pid, set the other pid to 0.
				hr = ms_pB2C2Adapter->GetAvControl()->
				         SetAudioVideoPIDs(lPartnerPid, (long)SParams.lVidPidArray[0]);

				if (FAILED (hr))
				{

					fprintf(stderr,"\nb2dump: Error: B2C2 Driver Data Ctrl."
					               " Interface SetAudioVideoPids method "
								   "failed, code: 0x%8X\n", hr);
					fflush(stderr);

					nRetVal = -1;

				}
				else
				{				
					m_lActivePid = SParams.lVidPidArray[0];
				}
			 }

			 else
			 {
				fprintf(stderr,"No Pids to add for Video stream.\n");
			 }

			 break;

		case AudioStream:

			if (SParams.bVerboseMode)
			{
				fprintf(stderr,"\nb2dump: AudioStream.");
			}

			 szSuffix = AUD_SUFFIX;

			 if (SParams.lTotalAudPidsSpecified)
			 {
				m_lActivePid = SParams.lAudPidArray[0];

				if (SParams.bVerboseMode)
				{
					fprintf(stderr,"\nb2dump: StartStream(AudioStream) -- Audio Pid %04X,"
				        " Video Pid %04X.\n", m_lActivePid, lPartnerPid);
				}

				// To set only one pid, set the other pid to 0.
				// NOTE: Audio and video pids need to be set simulteneously.
				// Here, we will set the audio pid only if there is no video pid.
				if (SParams.lTotalVidPidsSpecified == 0)
				{
					hr = ms_pB2C2Adapter->GetAvControl()->SetAudioVideoPIDs(m_lActivePid,0);

					if (FAILED (hr))
					{
						fprintf(stderr,"\nb2dump: Error: B2C2 Driver Data Ctrl."
									   " Interface SetAudioVideoPids method "
									   "failed, code: 0x%8X\n", hr);
						fflush(stderr);

						nRetVal = -1;

					}
				}

			 }
			 else
			 {
				fprintf(stderr,"No Pids to add for Audio stream.\n");
			 }

			 break;
	}

	// **********************************************************************
	// *** READ AND DUMP TO FILE
	// **********************************************************************

	// Open special device.
	if (!nRetVal)
	{
		strcat(caName, szSuffix);

		if (!(m_pStream = fopen (caName, "r")))
		{
			fprintf(stderr, "\nb2dump: Error: Cannot open special file \"%s\"\n",
		         	caName);
			fflush(stderr);

			nRetVal = -2;

		}
		else
		{
			if (SParams.bVerboseMode)
			{
				fprintf(stderr, "\nb2dump: Info: Opened special file \"%s\"\n", caName);
				fflush(stderr);
			}

			// Open dump file for writing.
        	strcpy(caName,SParams.szFilename);
        	strcat(caName,szSuffix);

			if (!(m_pDump = fopen(caName, "w")))
			{
				fprintf(stderr, "\nb2dump: Error: Cannot open dump file"
				                 "\"%s\"\n", caName);
				fflush(stderr);

				nRetVal = -3;

			}
			else
			{
				int nRet;
				BufferInfo bufInfo;

    			bufInfo.m_nSize = sizeof(BufferInfo);

				// ioctl returns nonzero in case of an error
				nRet = ioctl(fileno(m_pStream),B2C2AV_DRV_CMD_GET_BUF_INFO,&bufInfo);
			    if (nRet == 0)
				{

					if (SParams.bVerboseMode)
					{
						fprintf(stderr, "\nb2dump: Info: Got buffer info\n");
						fflush(stderr);
						fprintf (stderr, "bufInfo: size %d, pos %d, end %d. ioctl: %d\n", bufInfo.m_nSize, 
								bufInfo.m_ulBufferPosition, bufInfo.m_ulBufferEnd, nRet);
					}


					m_pucBaseAddr =
					   (unsigned char *) mmap(0, (unsigned int) bufInfo.m_ulBufferEnd,
												PROT_READ, MAP_FILE | MAP_PRIVATE,
												fileno(m_pStream), m_nOffset);

					if (m_pucBaseAddr == (void *)-1)
					{
						fprintf(stderr,"\nb2dump: Error: mmap call failed: %s\n", strerror(errno));
						fflush(stderr);
						nRetVal = -4;
					}
					else
					{
						if (SParams.bVerboseMode)
						{
							fprintf(stderr, "\nb2dump: mmap succeeded.\n");
							fflush(stderr);
						}
					}
				}
				else
				{
					fprintf (stderr, "ioctl error %d, B2C2AV_DRV_CMD_GET_BUF_INFO\n", nRet);
					fflush (stderr);
					nRetVal = -5;
				}
			}
		}
	}

	return nRetVal;
}

int CStreamRec::ProcessStream(bool bWriteFile, bool bVerboseMode)
{
	int nBytesRcv = 0;
	BufferInfo bufInfo;
	unsigned long ulCurPos;
	unsigned long ulBufSize;

    if (NULL != m_pStream)
	{
	    bufInfo.m_nSize = sizeof(BufferInfo);
	
		if (ioctl(fileno(m_pStream),B2C2AV_DRV_CMD_GET_BUF_INFO,&bufInfo))
		{
			fprintf(stderr,"b2dump: Info: GetBufferPos ioctl failed: %d.\n", errno);
		}
		else
		{
		    ulCurPos = bufInfo.m_ulBufferPosition;
			ulBufSize = bufInfo.m_ulBufferEnd;
	
			if (ulCurPos < 0 || ulCurPos > ulBufSize)
			{
				fprintf(stderr,"b2dump: Info: Bad buffer position pointer: %u, BufSize %ld\n",
				        ulCurPos, ulBufSize);
			}
			else
			{

				if (bVerboseMode)
				{
					fprintf(stderr,"b2dump: Info: ulCurPos: %u\n", ulCurPos);
				}
		
				if (ulCurPos > m_ulLastPos)
				{
					m_uLen = ulCurPos - m_ulLastPos;
					nBytesRcv += m_uLen;
					if (bWriteFile)
					{
						fwrite(m_pucBaseAddr + m_ulLastPos, 1, m_uLen, m_pDump);
					}
					m_ulLastPos = ulCurPos;
		
				}
				else if (ulCurPos < m_ulLastPos)
				{
					if (bVerboseMode)
					{
						fprintf(stderr,"b2dump: Info: Wrapped\n");
					}
		
					m_uLen = ulBufSize - m_ulLastPos;
					nBytesRcv += m_uLen;
					if (bWriteFile)
					{
						fwrite(m_pucBaseAddr + m_ulLastPos, 1, m_uLen, m_pDump);
					}

					m_uLen = ulCurPos;
					nBytesRcv += m_uLen;
					if (bWriteFile)
					{
						fwrite(m_pucBaseAddr, 1, m_uLen, m_pDump);
					}
					m_ulLastPos = ulCurPos;
		
				}
				else
				{
					if (bVerboseMode)
					{
						fprintf(stderr,"b2dump: Info: Not moving\n");
					}
				}
		
				fflush(stderr);
			}
		}
	}
	else
	{
		//fprintf (stderr, "b2dump: NULL stream!\n");
	}
	return nBytesRcv;
}

// **********************************************************************
// *** MAIN()
// **********************************************************************
int main(int argc, char * argv[])
{
	HRESULT					hr;
	tSParams				SParams;
	FILE *					m_pDump;

	// Need to be root to run; to test use effective user ID
	if (geteuid () != (uid_t) 0)
	{
		fprintf(stderr, "\nb2dump: Error: Must be root to run this command\n");
	
		return(1);	// *** FUNCTION EXIT POINT
	}

	// Initializations.
	SParams.lSysPidSize = MAX_B2DUMP_TS_PIDS;

	// Defaults for optional arguments.

	strcpy(SParams.szFilename, DUMP_FILENAME);
	SParams.lDumpDuration = 10;
	// Flags/counts for required arguments.

	SParams.bAdapterNameSet = FALSE;
	SParams.lTotalTSPidsSpecified = 0;
	SParams.lTotalVidPidsSpecified = 0;
	SParams.lTotalAudPidsSpecified = 0;

	// Zero statistics
	SParams.dwTotalTsBytes = 0;
	SParams.dwTotalAudBytes = 0;
	SParams.dwTotalVidBytes = 0;

	// Check command line arguments.
	int nTotalPids = CheckCommandLineArgs (argc, argv, &SParams);

	if (nTotalPids < 0)
	{
		return(1);	// *** FUNCTION EXIT POINT
	}

	// Make sure required flags have been set.

	if (SParams.bAdapterNameSet == FALSE)
	{
		fprintf(stderr, "\nb2dump: Error: Adapter name must be specified\n");
		fprintf(stderr, USAGE_MSG);
	
		return(1);	// *** FUNCTION EXIT POINT
	}

	if (nTotalPids < 1)
	{
		fprintf(stderr, "\nb2dump: Warning: No PID specified. PID must be set via SDK call to receive data\n");
	}

	if (SParams.lDumpDuration < 1)
	{
		fprintf(stderr, "\nb2dump: Error: Dump duration must be at least 1 sec.\n");
		fprintf(stderr, USAGE_MSG);
		fflush(stderr);
	
		return(1);	// *** FUNCTION EXIT POINT
	}

	CStreamRec TSStream(TsStream,&SParams);

	if (TSStream.InterfacesCreated())
	{
		CStreamRec AudStream(AudioStream,&SParams);
		CStreamRec VidStream(VideoStream,&SParams);

		// Note: video stream will set both Audio and Video pids, if specified, because they must
		// be set simulteneously. Otherwise, if there is only an Audio pid, it will be set by 
		// the Audio stream.
		if (TSStream.StartStream(SParams) || 
		    AudStream.StartStream(SParams) ||
		    VidStream.StartStream(SParams,&AudStream))
		{
			fprintf(stderr, "Could not start at least one of the streams.\n");

		}
		else
		{ 
			log = fopen(LOG_FILENAME, "w");

			int secs = SParams.lDumpDuration * 4; // Checked every 250ms

			for (int n = 0; n < secs; n++)
			{
				struct timeb StartTime, EndTime;

				ftime(&StartTime);  // Remember what time we started this itterration.

				SParams.dwTotalTsBytes += TSStream.ProcessStream(!SParams.bMonitorMode, SParams.bVerboseMode);
				SParams.dwTotalAudBytes += AudStream.ProcessStream(!SParams.bMonitorMode, SParams.bVerboseMode);
				SParams.dwTotalVidBytes += VidStream.ProcessStream(!SParams.bMonitorMode, SParams.bVerboseMode);

				fprintf(stderr,"b2dump: Total bytes: TS %ld Aud %ld Vid %ld\n",
				    SParams.dwTotalTsBytes,
					SParams.dwTotalAudBytes,
					SParams.dwTotalVidBytes);

				ftime(&EndTime);  // Get the ending time for the itteration.

				long lSeconds, lMilliSeconds;

				lSeconds = EndTime.time - StartTime.time;  // Compute ellapsed time in milliseconds.
				lMilliSeconds = (lSeconds * 1000L) + (EndTime.millitm - StartTime.millitm);

				if (lMilliSeconds < MINIMUM_ITTERATION_TIME)
				{
					usleep((MINIMUM_ITTERATION_TIME - lMilliSeconds)*1000L);
				}
			}
		}
	}
	else
	{
		fprintf(stderr,"\nb2dump: Error: B2C2 Driver Data Ctrl. Interface Initialize method failed, error: 0x%8.8x\n", hr);
	}

	fclose(log);

	return(0);
}

static long ConvertPid(char * strPid)
{
    long lRetVal;

	if (strlen (strPid) > 2 &&
		(strncmp(strPid, "0x", 2) == 0 ||
		 strncmp(strPid, "0X", 2) == 0))
	{
						// convert from hex

		lRetVal = strtoul((strPid + 2), NULL, 16);
	}
	else 
	{
		// assume decimal
		lRetVal = atoi(strPid);
	}

	return lRetVal;
}

// **********************************************************************
// *** CHECK COMMAND LINE ARGUMENTS
// **********************************************************************

int CheckCommandLineArgs(int argc, char * argv[], tSParams * pSParams)
{
	char		szTempBuf[1024];
	int			nPidCount = 0;
	long		lArgValueIndex;


	// Parse command line args, if any.  Eliminate program name itself which
	// counts as one arg.  All options are required unless specified otherwise
	//(see Usage message and comments).

	if (--argc > 0)
	{

		// Process and set values for specified flags.

		for (int ii = 1; ii <= argc; ii += (1 + lArgValueIndex))
		{
			// Most arguments require a value, so we set lArgValueIndex as 1.
			// For unary arguments such as "-st", lArgValueIndex will explicitly be
			// set to 0.

			lArgValueIndex = 1;
			if (pSParams->bVerboseMode)
			{
				fprintf(stderr, "Parsing %s\n", argv[ii]);
			}

			if (strcmp (argv[ii], "-f") == 0 || strcmp (argv[ii], "-F") == 0)
			{
				// Dump filename -- OPTIONAL

				strcpy(pSParams->szFilename, argv[ii + 1]);
			}
			else if (strcmp (argv[ii], "-a") == 0 || strcmp (argv[ii], "-A") == 0)
			{
				// Adapter Name for B2C2 MPEG2 card -- REQUIRED

				strcpy(pSParams->szAdapterName, argv[ii + 1]);
				pSParams->bAdapterNameSet = TRUE;
			}
			else if (strncmp (argv[ii], "-p", 2) == 0 ||
			         strncmp(argv[ii], "-P", 2) == 0)
			{
				// PID -- REQUIRED
				if (pSParams->bVerboseMode)
				{
					fprintf(stderr,"Parsing Pid Assignment: %c\n",
				        argv[ii][2]);
				}

                switch(argv[ii][2])
				{
				case 'v':
				case 'V':
					 if (pSParams->lTotalVidPidsSpecified == MAX_B2DUMP_AV_PIDS)
					 {
						 fprintf(stderr, "\nb2dump: Error: At most %ld Vid PIDs may be specified\n",
									 MAX_B2DUMP_AV_PIDS);

						 fprintf(stderr, USAGE_MSG);

						 return(-1);	// *** FUNCTION EXIT POINT
					 }

					 ++nPidCount;
					 pSParams->lVidPidArray[pSParams->lTotalVidPidsSpecified++] = 
						 ConvertPid(argv[ii + 1]);

				     break;

				case 'a':
				case 'A':
					 if (pSParams->lTotalAudPidsSpecified == MAX_B2DUMP_AV_PIDS)
					 {
						 fprintf(stderr, "\nb2dump: Error: At most %ld Aud PIDs may be specified\n",
									 MAX_B2DUMP_AV_PIDS);

						 fprintf(stderr, USAGE_MSG);

						 return(-1);	// *** FUNCTION EXIT POINT
					 }

					 ++nPidCount;
					 pSParams->lAudPidArray[pSParams->lTotalAudPidsSpecified++] = 
						 ConvertPid(argv[ii + 1]);

				     break;

				case 't':
				case 'T':
					 if (pSParams->lTotalTSPidsSpecified == pSParams->lSysPidSize)
					 {
						 fprintf(stderr, "\nb2dump: Error: At most %ld TS PIDs may be specified\n",
									 pSParams->lSysPidSize);

						 fprintf(stderr, USAGE_MSG);

						 return(-1);	// *** FUNCTION EXIT POINT
					 }

					 ++nPidCount;
					 pSParams->lTSPidArray[pSParams->lTotalTSPidsSpecified++] = 
						 ConvertPid(argv[ii + 1]);

				     break;

				default:
					// Catch-all error

					sprintf(szTempBuf, "\nb2dump: Error: Unrecognized option flag %s\n", argv[ii]);
					fprintf(stderr, szTempBuf);
					fprintf(stderr, USAGE_MSG);

					return(-1);	// *** FUNCTION EXIT POINT
				}

			}
			else if (strcmp (argv[ii], "-t") == 0 || strcmp (argv[ii], "-T") == 0)
			{
				// Duration(time in seconds) for dump to file -- OPTIONAL

				pSParams->lDumpDuration = atoi(argv[ii + 1]);
			}
			else if (strcmp (argv[ii], "-m") == 0 || strcmp (argv[ii], "-M") == 0)
			{
				// Monitor mode
				pSParams->bMonitorMode = TRUE;

				lArgValueIndex = 0;

			}
			else if (strcmp (argv[ii], "-v") == 0 || strcmp (argv[ii], "-V") == 0)
			{
				// Monitor mode
				pSParams->bVerboseMode = TRUE;

				lArgValueIndex = 0;

			}
			else
			{
				// Catch-all error

				fprintf(stderr,"\nb2dump: Error: Unrecognized option flag %s\n", argv[ii]);
				fprintf(stderr, USAGE_MSG);

				return(-1);	// *** FUNCTION EXIT POINT
			}
		}
	}

	if (pSParams->bVerboseMode)
	{
		fprintf(stderr,"\nb2dump:  %d PIDs set.\n", nPidCount);
	}

	return nPidCount;
}
